@@ -1,14 +1,14 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2013-2014 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2013-2015 -- leonerd@leonerd.org.uk
 
 package Devel::MAT;
 
 use strict;
 use warnings;
 
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 
 use Carp;
 use List::Util qw( pairs );
@@ -247,7 +247,15 @@ sub find_symbol
    my ( $sigil, $globname ) = $name =~ m/^([\$\@%&])(.*)$/ or
       croak "Could not parse sigil from $name";
 
-   my $glob = $self->find_glob( $globname );
+   my $stashvalue = $self->find_stashvalue( $globname );
+
+   # Perl 5.22 may take CODE shortcuts
+   if( $sigil eq '&' and $stashvalue->type eq "REF" ) {
+      return $stashvalue->rv;
+   }
+
+   $stashvalue->type eq "GLOB" or
+      croak "$globname is not a GLOB";
 
    my $slot = ( $sigil eq '$' ) ? "scalar" :
               ( $sigil eq '@' ) ? "array"  :
@@ -255,7 +263,7 @@ sub find_symbol
               ( $sigil eq '&' ) ? "code"   :
                                   die "ARGH"; # won't happen
 
-   my $sv = $glob->$slot or
+   my $sv = $stashvalue->$slot or
       croak "\*$globname has no $slot slot";
    return $sv;
 }
@@ -267,7 +275,7 @@ returning the C<GLOB> object if found.
 
 =cut
 
-sub find_glob
+sub find_stashvalue
 {
    my $self = shift;
    my ( $name ) = @_;
@@ -283,9 +291,21 @@ sub find_glob
       $stash = $self->dumpfile->defstash;
    }
 
-   my $gv = $stash->value( $shortname ) or
+   my $sv = $stash->value( $shortname ) or
       croak $stash->stashname . " has no symbol $shortname";
-   return $gv;
+   return $sv;
+}
+
+sub find_glob
+{
+   my $self = shift;
+   my ( $name ) = @_;
+
+   my $sv = $self->find_stashvalue( $name ) or return;
+   $sv->type eq "GLOB" or
+      croak "$name is not a GLOB";
+
+   return $sv;
 }
 
 =head1 AUTHOR
