@@ -8,7 +8,7 @@ package Devel::MAT::Context;
 use strict;
 use warnings;
 
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 
 use Carp;
 use Scalar::Util qw( weaken );
@@ -34,31 +34,38 @@ sub register_type
    *{"$_[0]::type"} = sub () { $typename };
 }
 
-sub _new
+sub new
 {
-   my $class = shift;
-   my ( $df ) = @_;
+   shift;
+   my ( $type, $df, $bytes, undef, $strs ) = @_;
 
-   my $self = bless {}, $class;
+   $types{$type} or croak "Cannot load unknown CTX type $type";
+
+   my $self = bless {}, $types{$type};
    weaken( $self->{df} = $df );
+
+   ( $self->{gimme}, $self->{line} ) = unpack "C $df->{uint_fmt}", $bytes;
+   ( $self->{file} ) = @$strs;
+
    return $self;
 }
 
-sub load
+sub load_v0_1
 {
    my $class = shift;
    my ( $type, $df ) = @_;
 
    $types{$type} or croak "Cannot load unknown CTX type $type";
 
-   my $self = $types{$type}->_new( $df );
+   my $self = bless {}, $types{$type};
+   weaken( $self->{df} = $df );
 
    # Standard fields all Contexts have
    $self->{gimme} = $df->_read_u8;
    $self->{file}  = $df->_read_str;
    $self->{line}  = $df->_read_uint;
 
-   $self->_load( $df );
+   $self->_load_v0_1( $df );
 
    return $self;
 }
@@ -101,7 +108,7 @@ sub location
 
 package Devel::MAT::Context::SUB;
 use base qw( Devel::MAT::Context );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 __PACKAGE__->register_type( 1 );
 
 =head1 Devel::MAT::Context::SUB
@@ -110,11 +117,25 @@ Represents a context which is a subroutine call.
 
 =cut
 
-sub _load
+sub load
+{
+   my $self = shift;
+   my ( $bytes, $ptrs, undef ) = @_;
+
+   my $df = $self->{df};
+
+   ( $self->{olddepth} ) = unpack "$df->{u32_fmt}", $bytes;
+
+   ( $self->{cv_at}, $self->{args_at} ) = @$ptrs;
+}
+
+sub _load_v0_1
 {
    my $self = shift;
    my ( $df ) = @_;
 
+   $self->{olddepth} = -1;
+
    $self->{cv_at}   = $df->_read_ptr;
    $self->{args_at} = $df->_read_ptr;
 }
@@ -127,14 +148,30 @@ Returns the CV which this call is to.
 
 Returns the arguments AV which represents the C<@_> argument array.
 
+=head2 $olddepth = $ctx->olddepth
+
+Returns the old depth of the context (that is, the depth the CV would be at
+after this context returns).
+
+=head2 $depth = $ctx->depth
+
+Returns the actual depth of the context. This is inferred at load time by
+considering the C<olddepth> of the next inner-nested call to the same CV, or
+from the actual C<depth> of the CV is no other call exists.
+
 =cut
 
 sub cv   { my $self = shift; return $self->{df}->sv_at( $self->{cv_at} ) }
 sub args { my $self = shift; return $self->{df}->sv_at( $self->{args_at} ) }
 
+sub olddepth { return $_[0]->{olddepth} }
+
+sub _set_depth { $_[0]->{depth} = $_[1] }
+sub depth      { return $_[0]->{depth} }
+
 package Devel::MAT::Context::TRY;
 use base qw( Devel::MAT::Context );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 __PACKAGE__->register_type( 2 );
 
 =head1 Devel::MAT::Context::TRY
@@ -143,11 +180,13 @@ Represents a context which is a block C<eval {}> call.
 
 =cut
 
-sub _load {}
+sub load {}
+
+sub _load_v0_1 {}
 
 package Devel::MAT::Context::EVAL;
 use base qw( Devel::MAT::Context );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 __PACKAGE__->register_type( 3 );
 
 =head1 Devel::MAT::Context::EVAL
@@ -156,6 +195,14 @@ Represents a context which is a string C<eval EXPR> call.
 
 =cut
 
+sub load
+{
+   my $self = shift;
+   my ( undef, $ptrs, undef ) = @_;
+
+   ( $self->{code_at} ) = @$ptrs;
+}
+
 sub _load
 {
    my $self = shift;
