@@ -8,7 +8,7 @@ package Devel::MAT::Dumper;
 use strict;
 use warnings;
 
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 
 use File::Basename qw( basename );
 
@@ -134,7 +134,7 @@ sub import
          my $old_DIE = $SIG{__DIE__};
          $SIG{__DIE__} = sub {
             local $SIG{__DIE__} = $old_DIE if defined $old_DIE;
-            die @_ if $^S;
+            return if $^S or !defined $^S; # only catch real process-fatal errors
 
             ( my $file = $dumpfile_name ) =~ s/NNN/$next_serial++/e;
             print STDERR "Dumping to $file because of DIE\n";
