@@ -1,14 +1,14 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2013-2014 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2013-2015 -- leonerd@leonerd.org.uk
 
 package Devel::MAT::Dumpfile;
 
 use strict;
 use warnings;
 
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 
 use Carp;
 use IO::Handle;   # ->read
@@ -173,7 +173,7 @@ sub load
 
    $self->_read_u8 == 0 or die "Cannot read $path - format version major unrecognised";
 
-   ( $self->{format_minor} = $self->_read_u8 ) <= 1 or
+   ( $self->{format_minor} = $self->_read_u8 ) <= 2 or
       die "Cannot read $path - format version minor unrecognised ($self->{format_minor})";
 
    if( $self->{format_minor} < 1 ) {
@@ -186,6 +186,12 @@ sub load
    my @sv_sizes = unpack "(a3)*", my $tmp = $self->_read( $n_types * 3 );
    $self->{sv_sizes} = [ map [ unpack "C C C", $_ ], @sv_sizes ];
 
+   if( $self->{format_minor} >= 2 ) {
+      my $n_ctxs = $self->_read_u8;
+      my @ctx_sizes = unpack "(a3)*", my $tmp = $self->_read( $n_ctxs * 3 );
+      $self->{ctx_sizes} = [ map [ unpack "C C C", $_ ], @ctx_sizes ];
+   }
+
    # Roots
    foreach (qw( undef yes no )) {
       my $addr = $self->{"${_}_at"} = $self->_read_ptr;
@@ -270,6 +276,20 @@ sub _fixup
          $count, $heap_total, 100*$count / $heap_total ) if $progress and ($count % 5000) == 0;
    }
 
+   # Walk the SUB contexts setting their true depth
+   if( $self->{format_minor} >= 2 ) {
+      my %prev_depth_by_cvaddr;
+
+      foreach my $ctx ( @{ $self->{contexts} } ) {
+         next unless $ctx->type eq "SUB";
+
+         my $cvaddr = $ctx->{cv_at};
+         $ctx->_set_depth( $prev_depth_by_cvaddr{$cvaddr} // $ctx->cv->depth );
+
+         $prev_depth_by_cvaddr{$cvaddr} = $ctx->olddepth;
+      }
+   }
+
    return $self;
 }
 
@@ -341,6 +361,17 @@ sub _read_str
    return $self->_read($len);
 }
 
+sub _read_bytesptrsstrs
+{
+   my $self = shift;
+   my ( $nbytes, $nptrs, $nstrs ) = @_;
+
+   return
+      ( $nbytes ? $self->_read( $nbytes ) : "" ),
+      ( $nptrs  ? [ $self->_read_ptrs( $nptrs ) ] : undef ),
+      ( $nstrs  ? [ map { $self->_read_str } 1 .. $nstrs ] : undef );
+}
+
 sub _read_sv
 {
    my $self = shift;
@@ -379,21 +410,13 @@ sub _read_sv
       my $pos = tell $self->{fh};
 
       # First read the "common" header
-      my ( $nbytes, $nptrs, $nstrs ) = @{ $self->{sv_sizes}[0] };
-
       my $sv = Devel::MAT::SV->new( $type, $self,
-         ( $nbytes ? $self->_read( $nbytes ) : "" ), # common header
-         ( $nptrs  ? [ $self->_read_ptrs( $nptrs ) ] : undef ),
-         ( $nstrs  ? [ map { $self->_read_str } 1 .. $nstrs ] : undef ),
+         $self->_read_bytesptrsstrs( @{ $self->{sv_sizes}[0] } )
       );
 
       # Then the SV header
-      ( $nbytes, $nptrs, $nstrs ) = @{ $self->{sv_sizes}[$type] };
-
       $sv->load(
-         ( $nbytes ? $self->_read( $nbytes ) : "" ), # header
-         ( $nptrs  ? [ $self->_read_ptrs( $nptrs ) ] : undef ),
-         ( $nstrs  ? [ map { $self->_read_str } 1 .. $nstrs ] : undef ),
+         $self->_read_bytesptrsstrs( @{ $self->{sv_sizes}[$type] } )
       );
 
       return $sv;
@@ -407,7 +430,20 @@ sub _read_ctx
    my $type = $self->_read_u8;
    return if !$type;
 
-   return Devel::MAT::Context->load( $type, $self );
+   if( $self->{format_minor} >= 2 ) {
+      my $ctx = Devel::MAT::Context->new( $type, $self,
+         $self->_read_bytesptrsstrs( @{ $self->{ctx_sizes}[0] } )
+      );
+
+      $ctx->load(
+         $self->_read_bytesptrsstrs( @{ $self->{ctx_sizes}[$type] } )
+      );
+
+      return $ctx;
+   }
+   else {
+      return Devel::MAT::Context->load_v0_1( $type, $self );
+   }
 }
 
 =head1 METHODS
