@@ -10,7 +10,7 @@ use warnings;
 use feature qw( switch );
 no if $] >= 5.017011, warnings => 'experimental::smartmatch';
 
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 
 use Carp;
 use Scalar::Util qw( weaken );
@@ -332,7 +332,7 @@ boolean true and false. They are
 
 package Devel::MAT::SV::Immortal;
 use base qw( Devel::MAT::SV );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 use constant immortal => 1;
 sub new {
    my $class = shift;
@@ -345,13 +345,13 @@ sub _outrefs { () }
 
 package Devel::MAT::SV::UNDEF;
 use base qw( Devel::MAT::SV::Immortal );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 sub desc { "UNDEF" }
 sub type { "UNDEF" }
 
 package Devel::MAT::SV::YES;
 use base qw( Devel::MAT::SV::Immortal );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 sub desc { "YES" }
 sub type { "SCALAR" }
 
@@ -366,7 +366,7 @@ sub name {}
 
 package Devel::MAT::SV::NO;
 use base qw( Devel::MAT::SV::Immortal );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 sub desc { "NO" }
 sub type { "SCALAR" }
 
@@ -381,7 +381,7 @@ sub name {}
 
 package Devel::MAT::SV::Unknown;
 use base qw( Devel::MAT::SV );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 __PACKAGE__->register_type( 0xff );
 
 sub desc { "UNKNOWN" }
@@ -390,7 +390,7 @@ sub _outrefs {}
 
 package Devel::MAT::SV::GLOB;
 use base qw( Devel::MAT::SV );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 __PACKAGE__->register_type( 1 );
 
 =head1 Devel::MAT::SV::GLOB
@@ -521,7 +521,7 @@ sub _outrefs
 
 package Devel::MAT::SV::SCALAR;
 use base qw( Devel::MAT::SV );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 __PACKAGE__->register_type( 2 );
 
 =head1 Devel::MAT::SV::SCALAR
@@ -616,6 +616,8 @@ sub qq_pv
 
 Returns the stash of the SCALAR, if it is an 'C<our>' variable.
 
+After perl 5.20 this is no longer used, and will return C<undef>.
+
 =cut
 
 sub ourstash { my $self = shift; return $self->df->sv_at( $self->ourstash_at ) }
@@ -651,7 +653,7 @@ sub _outrefs
 
 package Devel::MAT::SV::REF;
 use base qw( Devel::MAT::SV );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 __PACKAGE__->register_type( 3 );
 
 =head1 Devel::MAT::SV::REF
@@ -722,7 +724,7 @@ sub _outrefs
 
 package Devel::MAT::SV::ARRAY;
 use base qw( Devel::MAT::SV );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 __PACKAGE__->register_type( 4 );
 
 =head1 Devel::MAT::SV::ARRAY
@@ -834,7 +836,7 @@ sub _outrefs
 package Devel::MAT::SV::PADLIST;
 # Synthetic type
 use base qw( Devel::MAT::SV::ARRAY );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 use constant type => "PADLIST";
 
 =head1 Devel::MAT::SV::PADLIST
@@ -870,7 +872,7 @@ sub _outrefs
 package Devel::MAT::SV::PADNAMES;
 # Synthetic type
 use base qw( Devel::MAT::SV::ARRAY );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 use constant type => "PADNAMES";
 
 =head1 Devel::MAT::SV::PADNAMES
@@ -896,6 +898,27 @@ sub padname
    return $namepv->pv;
 }
 
+=head2 $padix = $padnames->padix_from_padname( $padname )
+
+Returns the index of the lexical with the given name, or C<undef>
+
+=cut
+
+sub padix_from_padname
+{
+   my $self = shift;
+   my ( $padname ) = @_;
+
+   foreach my $padix ( 1 .. scalar( $self->elems ) - 1 ) {
+      my $namepv;
+      return $padix if $namepv = $self->elem( $padix ) and
+                       $namepv->type eq "SCALAR" and
+                       $namepv->pv eq $padname;
+   }
+
+   return undef;
+}
+
 sub desc
 {
    my $self = shift;
@@ -921,7 +944,7 @@ sub _outrefs
 package Devel::MAT::SV::PAD;
 # Synthetic type
 use base qw( Devel::MAT::SV::ARRAY );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 use constant type => "PAD";
 
 use List::Util qw( pairmap );
@@ -953,8 +976,8 @@ sub lexvars
 
    my @svs = $self->elems;
    return map {
-      my $name = $padcv->padname( $_ );
-      $name ? ( $name => $svs[$_] ) : ()
+      my $padname = $padcv->padname( $_ );
+      $padname ? ( $padname->name => $svs[$_] ) : ()
    } 1 .. $#svs;
 }
 
@@ -970,7 +993,8 @@ sub _outrefs
       '+the @_ av' => $svs[0],
       map {
          my $sv = $svs[$_];
-         my $name = $padcv->padname( $_ );
+         my $padname = $padcv->padname( $_ );
+         my $name = $padname ? $padname->name : undef;
          $name ? ( $direct_or_rv->( "the lexical $name" => $sv ) )
                : ( $direct_or_rv->( "elem [$_]" => $sv ) )
       } 1 .. $#svs,
@@ -979,7 +1003,7 @@ sub _outrefs
 
 package Devel::MAT::SV::HASH;
 use base qw( Devel::MAT::SV );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 __PACKAGE__->register_type( 5 );
 
 =head1 Devel::MAT::SV::HASH
@@ -1102,7 +1126,7 @@ sub _outrefs
 
 package Devel::MAT::SV::STASH;
 use base qw( Devel::MAT::SV::HASH );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 __PACKAGE__->register_type( 6 );
 
 =head1 Devel::MAT::SV::STASH
@@ -1150,6 +1174,34 @@ sub mro_linearcurrent { my $self = shift; return $self->df->sv_at( $self->{mro_l
 sub mro_nextmethod    { my $self = shift; return $self->df->sv_at( $self->{mro_nextmethod_at} ) }
 sub mro_isa           { my $self = shift; return $self->df->sv_at( $self->{mro_isa_at} ) }
 
+=head2 $cv = $stash->value_code( $key )
+
+Returns the CODE associated with the given symbol name, if it exists, or
+C<undef> if not. This is roughly equivalent to
+
+   $cv = $stash->value( $key )->code
+
+Except that it is aware of the direct reference to CVs that perl 5.22 will
+optimise for. This method should be used in preference to the above construct.
+
+=cut
+
+sub value_code
+{
+   my $self = shift;
+   my ( $key ) = @_;
+
+   my $sv = $self->value( $key ) or return undef;
+   if( $sv->type eq "GLOB" ) {
+      return $sv->code;
+   }
+   elsif( $sv->type eq "REF" ) {
+      return $sv->rv;
+   }
+
+   die "TODO: value_code on non-GLOB, non-REF ${\ $sv->desc }";
+}
+
 =head2 $name = $stash->stashname
 
 Returns the name of the stash
@@ -1182,11 +1234,16 @@ sub _outrefs
 
 package Devel::MAT::SV::CODE;
 use base qw( Devel::MAT::SV );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 __PACKAGE__->register_type( 7 );
 
+use Carp;
+
 use List::MoreUtils qw( uniq );
 
+use Struct::Dumb qw( readonly_struct );
+readonly_struct Padname => [qw( name ourstash )];
+
 =head1 Devel::MAT::SV::CODE
 
 Represents a function or closure; an SV of type C<SVt_PVCV>.
@@ -1199,23 +1256,28 @@ sub load
    my ( $header, $ptrs, $strs ) = @_;
    my $df = $self->df;
 
-   my ( $line, $flags, $oproot ) =
-      unpack "$df->{uint_fmt} C $df->{ptr_fmt}", $header;
+   my ( $line, $flags, $oproot, $depth ) =
+      unpack "$df->{uint_fmt} C $df->{ptr_fmt} $df->{u32_fmt}", $header;
 
-   $self->_set_code_fields( $line, $flags, $oproot,
+   defined $depth or $depth = -1;
+
+   $self->_set_code_fields( $line, $flags, $oproot, $depth,
       @{$ptrs}[0, 2..4], # STASH, OUTSIDE, PADLIST, CONSTVAL
-      $strs->[0],        # FILE
+      @{$strs}[0, 1],    # FILE, NAME
    );
    $self->_set_glob_at( $ptrs->[1] );
 
+   # After perl 5.20 individual padname structs are no longer arena-allocated
+   $self->{padnames} = [] if $df->{perlver} > ( ( 5 << 24 ) | ( 20 << 16 ) | 0xffff );
+
    while( my $type = $df->_read_u8 ) {
       given( $type ) {
          when( 1 ) { push @{ $self->{consts_at} }, $df->_read_ptr }
          when( 2 ) { push @{ $self->{constix} }, $df->_read_uint }
          when( 3 ) { push @{ $self->{gvs_at} }, $df->_read_ptr }
          when( 4 ) { push @{ $self->{gvix} }, $df->_read_uint }
-         when( 5 ) { # ignore - used to be padname
-                     $df->_read_uint; $df->_read_str }
+         when( 5 ) { my $padix = $df->_read_uint;
+                     $self->{padnames}[$padix] = _load_padname( $df ); }
          when( 6 ) { # ignore - used to be padsvs_at
                      $df->_read_uint; $df->_read_uint; $df->_read_ptr; }
          when( 7 ) { $self->_set_padnames_at( $df->_read_ptr ); }
@@ -1226,6 +1288,13 @@ sub load
    }
 }
 
+sub _load_padname
+{
+   my ( $df ) = @_;
+
+   return Padname( $df->_read_str, $df->_read_ptr );
+}
+
 sub _fixup
 {
    my $self = shift;
@@ -1234,8 +1303,10 @@ sub _fixup
    my $df = $self->df;
 
    my $padlist = $self->padlist;
-   bless $padlist, "Devel::MAT::SV::PADLIST" if $padlist;
-   $padlist->_set_padcv_at( $self->addr ) if $padlist;
+   if( $padlist ) {
+      bless $padlist, "Devel::MAT::SV::PADLIST";
+      $padlist->_set_padcv_at( $self->addr );
+   }
 
    my $padnames;
    my @pads;
@@ -1254,8 +1325,19 @@ sub _fixup
       $self->_set_padnames_at( $padnames->addr );
    }
 
-   bless $padnames, "Devel::MAT::SV::PADNAMES";
-   $padnames->_set_padcv_at( $self->addr );
+   if( $padnames ) {
+      bless $padnames, "Devel::MAT::SV::PADNAMES";
+      $padnames->_set_padcv_at( $self->addr );
+
+      $self->{padnames} = \my @padnames;
+
+      foreach my $padix ( 1 .. $padnames->elems - 1 ) {
+         my $padnamesv = $padnames->elem( $padix ) or next;
+         $padnamesv->immortal and next; # UNDEF
+
+         $padnames[$padix] = Padname( $padnamesv->pv, $padnamesv->ourstash );
+      }
+   }
 
    foreach my $pad ( @pads ) {
       next unless $pad;
@@ -1302,8 +1384,10 @@ sub _fixup
 
 =head2 $addr = $cv->oproot
 
-Returns the stash, glob, filename, line number, scope, padlist, constant value
-or oproot of the code.
+=head2 $depth = $cv->depth
+
+Returns the stash, glob, filename, line number, scope, padlist, constant value,
+oproot or depth of the code.
 
 =cut
 
@@ -1313,7 +1397,7 @@ sub glob     { my $self = shift; return $self->df->sv_at( $self->glob_at ) }
 sub scope    { my $self = shift; return $self->df->sv_at( $self->outside_at ) }
 sub padlist  { my $self = shift; return $self->df->sv_at( $self->padlist_at ) }
 sub constval { my $self = shift; return $self->df->sv_at( $self->constval_at ) }
-# XS accessor: oproot
+# XS accessors: oproot, depth
 
 =head2 $location = $cv->location
 
@@ -1340,8 +1424,10 @@ sub location
 
 =head2 $rc = $cv->is_cvgv_rc
 
-Returns the C<CvCLONE()>, C<CvCLONED()>, C<CvISXSUB()>, C<CvWEAKOUTSIDE()> and
-C<CvCVGV_RC()> flags.
+=head2 $lexical = $cv->is_lexical
+
+Returns the C<CvCLONE()>, C<CvCLONED()>, C<CvISXSUB()>, C<CvWEAKOUTSIDE()>,
+C<CvCVGV_RC()> and C<CvLEXICAL()> flags.
 
 =cut
 
@@ -1390,14 +1476,26 @@ sub stashname { my $self = shift; return $self->stash ? $self->stash->stashname
 sub name
 {
    my $self = shift;
-   return unless my $glob_at = $self->glob_at;
-   return '&' . $self->df->sv_at( $glob_at )->stashname;
+
+   # CvLEXICALs or CVs with non-reified CvGVs may still have a hekname
+   if( defined( my $hekname = $self->hekname ) ) {
+      return '&' . $self->stashname . "::" . $hekname;
+   }
+   elsif( my $glob = $self->glob ) {
+      return '&' . $glob->stashname;
+   }
+
+   return undef;
 }
 
-=head2 $name = $cv->padname( $padix )
+=head2 $padname = $cv->padname( $padix )
 
 Returns the name of the $padix'th lexical variable, or C<undef> if it doesn't
-have a name
+have a name.
+
+The returned padname is a structure of the following fields:
+
+ $name = $padname->name
 
 =cut
 
@@ -1406,19 +1504,47 @@ sub padname
    my $self = shift;
    my ( $padix ) = @_;
 
-   return $self->padnames->padname( $padix );
+   return $self->{padnames}[$padix];
+}
+
+=head2 $padix = $cv->padix_from_padname( $padname )
+
+Returns the index of the first lexical variable with the given pad name, or
+C<undef> if one does not exist.
+
+=cut
+
+sub padix_from_padname
+{
+   my $self = shift;
+   my ( $padname ) = @_;
+
+   my $padnames = $self->{padnames};
+
+   foreach my $padix ( 1 .. $#$padnames ) {
+      return $padix if defined $padnames->[$padix] and
+                       $padnames->[$padix]->name eq $padname;
+   }
+
+   return undef;
 }
 
 =head2 $padnames = $cv->padnames
 
 Returns the AV reference directly which stores the pad names.
 
+After perl version 5.20, this is no longer used directly and will return
+C<undef>. The individual pad names themselves can still be found via the
+C<padname> method.
+
 =cut
 
 sub padnames
 {
    my $self = shift;
-   return $self->df->sv_at( $self->padnames_at );
+
+   return $self->df->sv_at( $self->padnames_at or return undef )
+      // croak "${\ $self->desc } PADNAMES is not accessible";
 }
 
 =head2 @pads = $cv->pads
@@ -1435,7 +1561,7 @@ sub pads
 
 =head2 $pad = $cv->pad( $depth )
 
-Returns the PAD at the given depth
+Returns the PAD at the given depth (given by 1-based index).
 
 =cut
 
@@ -1443,7 +1569,7 @@ sub pad
 {
    my $self = shift;
    my ( $depth ) = @_;
-   return $self->{pads} ? $self->{pads}[$depth] : undef;
+   return $self->{pads} ? $self->{pads}[$depth-1] : undef;
 }
 
 sub desc
@@ -1504,7 +1630,7 @@ sub _outrefs
 
 package Devel::MAT::SV::IO;
 use base qw( Devel::MAT::SV );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 __PACKAGE__->register_type( 8 );
 
 sub load
@@ -1534,7 +1660,7 @@ sub _outrefs
 
 package Devel::MAT::SV::LVALUE;
 use base qw( Devel::MAT::SV );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 __PACKAGE__->register_type( 9 );
 
 sub load
@@ -1567,7 +1693,7 @@ sub _outrefs
 
 package Devel::MAT::SV::REGEXP;
 use base qw( Devel::MAT::SV );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 __PACKAGE__->register_type( 10 );
 
 sub load {}
@@ -1578,7 +1704,7 @@ sub _outrefs { () }
 
 package Devel::MAT::SV::FORMAT;
 use base qw( Devel::MAT::SV );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 __PACKAGE__->register_type( 11 );
 
 sub load {}
@@ -1589,7 +1715,7 @@ sub _outrefs { () }
 
 package Devel::MAT::SV::INVLIST;
 use base qw( Devel::MAT::SV );
-our $VERSION = '0.20';
+our $VERSION = '0.21';
 __PACKAGE__->register_type( 12 );
 
 sub load {}
